#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
  
#define SOFTWARE_VERSION   	   "bmp2c/0.1 jun2006"

struct biInfo {
   unsigned long bfOffBits; 	/* 4 bytes */
   unsigned long biWidth;  	/* 4 bytes */
   unsigned long biHeight; 	/* 4 bytes */
   unsigned int biBitCount;	/* 2 bytes */	
   struct stat biFileInfo;
};

struct biFileAccess {
   FILE* biInpFile;
   FILE* biOutFile;
};

typedef struct biInfo biInfo_t;
typedef struct biFileAccess biFileAccess_t;

/* ----- global Vars. ----- */

/* ------ local Vars. ----- */

/* ------- forwards ------- */
static int open_f_stdin( biFileAccess_t*, char*, biInfo_t* );
static void print_pict_info( biInfo_t* );
static void usage ( char* );
static void help  ( char* );

/* static */ int 
biFileValid(unsigned char* biFileHeader, biInfo_t* chk_f_info)
{
   unsigned char i;
   unsigned long biFileSize;
   
   /* check 'BM' */
   
   if ( (biFileHeader[0] & ~(0x42)) || (biFileHeader[1] & ~(0x4d)) )
   {
   	 (void) fprintf( stderr, "error: invalid file type\r\n");
	 return -1;
   }
   
   /* check file size */
   
   biFileSize = 0;
   for(i=4; i>0; i--)
   {
   	 biFileSize |= biFileHeader[i+1] << ((i-1)<<3);
   }
   if (biFileSize != chk_f_info->biFileInfo.st_size)
   {
   	 (void) fprintf( stderr, "error: real file size and file header info. mismatch\r\n");
	 return -1;
   }
   
   /* record picture data position */
   
   chk_f_info->bfOffBits = 0;
   for(i=4; i>0; i--)
   {
   	 chk_f_info->bfOffBits |= biFileHeader[i+9] << ((i-1)<<3);
   }
   
   return 0;   /* no error */
}

/* static */ int
biExtractImageInfo(unsigned char* biBitmapHeader, biInfo_t* bmp_f_info)
{
   unsigned char i;
   unsigned long biCompression;

   /* get biWidth    (4 bytes)
   	     biHeight 	 (4 bytes)
	     biBitCount (2 bytes) */
   
   bmp_f_info->biWidth = 0;
   for(i=4; i>0; i--)
   {
   	 bmp_f_info->biWidth |= biBitmapHeader[i+3] << ((i-1)<<3);
   }
   
   bmp_f_info-> biHeight = 0;
   for(i=4; i>0; i--)
   {
   	 bmp_f_info->biHeight |= biBitmapHeader[i+7] << ((i-1)<<3);
   }
   
   /* 'biPlanes' (2 bytes) skipped */

   bmp_f_info->biBitCount = 0;
   for(i=2; i>0; i--)
   {
   	 bmp_f_info->biBitCount |= biBitmapHeader[i+13] << ((i-1)<<3);
   }
   if(bmp_f_info->biBitCount != 24)
   {
   	 (void) fprintf( stderr, "error: only RGB24 pictures are supported in this version\r\n" );
	 return -1;
   }

   /* check for no compression 'biCompression' 4 bytes */
   
   biCompression = 0;
   for(i=4; i>0; i--)
   {
   	 biCompression |= biBitmapHeader[i+15] << ((i-1)<<3);
   }
   if( biCompression )
   {
   	 (void) fprintf( stderr, "error: only uncompressed pictures are supported in this version\r\n" );
      return -1;
   }

   return 0;   /* no error */
}

int main(int argc , char ** argv)
{
   char* argv0;
   int argn = 1;
   
   biInfo_t f_info;
   biFileAccess_t f_access;
   unsigned long biSize;

   unsigned char i;

   unsigned char caught_dat[1024];        /* 1ko */
   long acq_length,f_length;
   int line_ind;
   
//   FILE* f_stdin;
 
  /* ------------------------------------------------------------------------- */
  /* ------------------------------ Parse args. ------------------------------ */
  /* ------------------------------------------------------------------------- */ 

  argv0 = argv[0];
   
  int args_dat_array[3] = {-1, -1, -1};             /* flags dat. pos. : i n o */

  while ( argn < argc )
  {
      if ( strcmp( argv[argn], "--version" ) == 0 )
         {
            (void) printf( "%s\r\n", SOFTWARE_VERSION);
            exit( 0 );
         }
      else if ( strcmp( argv[argn], "--help" ) == 0 )
         {
            help(argv0);
            exit( 0 );
         }
      else if ( strcmp( argv[argn], "-i" ) == 0 )
         {    
            ++argn;
            if (argn == argc) usage(argv0);
		  args_dat_array[0] = argn;
            ++argn;
         }    
      else if ( strcmp( argv[argn], "-n" ) == 0 )
         {
            ++argn;
            if (argn == argc) usage(argv0);
		  args_dat_array[1] = argn;
            ++argn;
         }
      else if ( strcmp( argv[argn], "-o" ) == 0 )
         {
            ++argn;
            if (argn == argc) usage(argv0);
		  args_dat_array[2] = argn;
            ++argn;
         }
      else
         usage(argv0);
  }

/* args debug */
/*
   printf("i: %d\r\n", args_dat_array[0]);
   printf("n: %d\r\n", args_dat_array[1]);
   printf("o: %d\r\n", args_dat_array[2]);

   if ( args_dat_array[0] > 0)
   {
   	 printf("i dat.: %s\r\n", argv[args_dat_array[0]]);
   }
   if ( args_dat_array[1] > 0)
   {
   	 printf("n dat.: %s\r\n", argv[args_dat_array[1]]);
   }
   if ( args_dat_array[2] > 0)
   {
   	 printf("o dat.: %s\r\n", argv[args_dat_array[2]]);
   }

   exit(-1);
*/

   if (args_dat_array[0] != -1)  /* -i option */
   {
      if (open_f_stdin(&f_access, argv[args_dat_array[0]], &f_info) < 0) exit(-1);
   }else{
      if (open_f_stdin(&f_access, "in.bmp" , &f_info) < 0) exit(-1);
   }
   
   /* Read the BITMAPFILEHEADER (14 bytes) */
   acq_length = fread(caught_dat,1,14,f_access.biInpFile);
   
   if (acq_length < 14)
   { 
   	 printf("error: unexpected EOF\r\n");
	 fclose ( f_access.biInpFile );	 
	 exit( -1 );
   }
   
   if (biFileValid(caught_dat, &f_info))
   {
      printf("assert [biFileValid]\r\n");
      fclose ( f_access.biInpFile );
      exit( -1 );
   }

   /* Extract bitmap info. in the BITMAPINFOHEADER */
   acq_length = fread(caught_dat,1,20,f_access.biInpFile);

   if (acq_length < 20)
   { 
   	 printf("error: unexpected EOF\r\n");
	 fclose ( f_access.biInpFile );
	 exit( -1 );
   }
   
   if (biExtractImageInfo(caught_dat, &f_info))
   {
      fclose ( f_access.biInpFile );
      exit( -1 );
   }

   print_pict_info(&f_info);

   /* extract pixels data */

   unsigned char padding;
   unsigned long fWidthLength;
   unsigned char* biLineBuffer;

   padding = (((f_info.biWidth*3) % 4) ? 4-((f_info.biWidth*3) % 4) : 0);

   /* printf(">>%d\r\n", (f_info.biWidth*3) + padding); */

   fWidthLength = (f_info.biWidth*3) + padding;

   biLineBuffer = (char*) malloc (fWidthLength);
   if (biLineBuffer == NULL)
   {
      printf("memory allocation failed [biLineBuffer]\r\n");
      fclose ( f_access.biInpFile );
      exit( -1 );
   }

   if( !(((fWidthLength * f_info.biHeight) + f_info.bfOffBits) - f_info.biWidth) )
   {
      printf("error: real file size and calculated file size mismatch\r\n");
      fclose ( f_access.biInpFile );
      exit( -1 );
   }

   if (args_dat_array[1] != -1)  /* -n option */
   {
      //if (open_f_stdout(&f_access, argv[args_dat_array[1]], &f_info) < 0) exit(-1);
      //printf("n dat.: %s\r\n", argv[args_dat_array[1]]);
   }else{
      //printf("no n dat.");
   }

   unsigned long k,l;
   unsigned long pixels=0;

   printf("\r\nunsigned char picture_dat[]={\r\n");    /* debug only */

   fseek(f_access.biInpFile, -fWidthLength, SEEK_END);

   for(k=0; k<f_info.biHeight-1; k++)
   {

      fread(biLineBuffer, fWidthLength, 1, f_access.biInpFile);
      for(l=(f_info.biWidth*3); l>0; l--)
      {
         printf("0x%02x,", biLineBuffer[l-1]);
         pixels++;
      }
      printf("\r\n");
      fseek(f_access.biInpFile, -(fWidthLength*2), SEEK_CUR);

   }

   fread(biLineBuffer, fWidthLength, 1, f_access.biInpFile);
   for(l=(f_info.biWidth*3); l>0; l--)
   {
      if(l==1)
      {
         printf("0x%02x};\r\n", biLineBuffer[l-1]);
         pixels++;
      }else{
         printf("0x%02x,", biLineBuffer[l-1]);
         pixels++;
      }
   }

/*
   fseek(f_access.biInpFile, -fWidthLength, SEEK_END);

   for(k=0; k<f_info.biHeight-1; k++)
   {

      fread(biLineBuffer, fWidthLength, 1, f_access.biInpFile);
      for(l=(f_info.biWidth*3); l>0; l-=3)
      {
         printf("0x%02x,", biLineBuffer[l-3]);
         printf("0x%02x,", biLineBuffer[l-2]);
         printf("0x%02x,", biLineBuffer[l-1]); 
         pixels+=3;
      }
      printf("\r\n");
      fseek(f_access.biInpFile, -(fWidthLength*2), SEEK_CUR);
   }

   fread(biLineBuffer, fWidthLength, 1, f_access.biInpFile);
   for(l=(f_info.biWidth*3); l>0; l-=3)
   {
      if(l==3)
      {
         printf("0x%02x,", biLineBuffer[l-3]);
         printf("0x%02x,", biLineBuffer[l-2]);
         printf("0x%02x};\r\n", biLineBuffer[l-1]);
         pixels+=3;
      }else{
         printf("0x%02x,", biLineBuffer[l-3]);
         printf("0x%02x,", biLineBuffer[l-2]);
         printf("0x%02x,", biLineBuffer[l-1]);
         pixels+=3;
      }
   }
*/
   printf("\r\n/* pixels count : %d */\r\n", pixels);

   free (biLineBuffer);

   /***********************/

  // fflush( f_stdout );
   fclose( f_access.biInpFile );

   return 0;
}


static int
open_f_stdin (biFileAccess_t* f_in_access, char* f_in_name, biInfo_t* f_in_pict_info )
{

      if (stat(f_in_name, &f_in_pict_info->biFileInfo) == -1)
      {
          if (errno == ENOENT)
          {
            (void) fprintf( stderr, "error: '%s' does not exist\r\n", f_in_name );
          }else{
            (void) fprintf( stderr, "error: stat() fails for '%s'\r\n", f_in_name );
          }
          return -1;
      }

      /* checking S_ISLNK(m) is not required since 'stat' follows sym. links */
      if (!S_ISREG(f_in_pict_info->biFileInfo.st_mode))
      {
         (void) fprintf( stderr, "error: '%s' is a not a regular file (or a sym. link)\r\n", f_in_name);
         return -1;
      }

      if ((f_in_access->biInpFile = fopen(f_in_name, "r")) == NULL)
      {
         (void) fprintf( stderr, "error: '%s' open failed\r\n", f_in_name );
         return -1;
      }
      
      return 0;   /* no error */
}

static void
print_pict_info( biInfo_t* pict_info )
{
   unsigned char p_info_padding;

   /* function core */

   printf("/*\r\n");
   printf("file size (bmp)   : %ld bytes\r\n", pict_info->biFileInfo.st_size);
   printf("width             : %ld pixel(s)\r\n", pict_info->biWidth);
   printf("height            : %ld pixel(s)\r\n", pict_info->biHeight);
   printf("bit(s) per pixel  : %d\r\n", pict_info->biBitCount);
   printf("data array offset : 0x%lx\r\n", pict_info->bfOffBits);

   /****************/

   /* file size calculation */
   printf("\r\n");
   /* printf("%f\r\n", ((pict_info->biWidth*3.00)/(4.00))); */
   p_info_padding = 4 - ((pict_info->biWidth*3) % 4);
   if(p_info_padding < 4)
   {
      printf("padding           : %d byte(s)\r\n", p_info_padding);  
   }else{
      printf("no padding\r\n");
      p_info_padding = 0;
   }

   printf("calc. file size   : %ld bytes\r\n", (((pict_info->biWidth*3)+p_info_padding) * pict_info->biHeight) + pict_info->bfOffBits);
   printf("*/\r\n");
}

static void
usage ( char* use_argv0 )
{
   (void) fprintf( stderr, "usage: %s [-i input_file_name] [-n array_name] [-o output_file_name]\r\n", use_argv0 );
   exit(1);
}

static void
help ( char* hlp_argv0 )
{
   (void) fprintf( stdout, "usage: %s [-i input_file_name] [-n array_name] [-o output_file_name]\r\n\r\n", hlp_argv0);
   (void) fprintf( stdout, "Options:\r\n  -i FILE\t\tspecifie an input file name\r\n         \t\t(\"in.bmp\" if not specified)\r\n  -n ARRAY_NAME\t\traw data array name(default: \'picture_dat[]\')\r\n  -o FILE\t\tspecifie an output file name\r\n         \t\t(\"out.raw\" if not specified)\r\n      --help     aide-mmoire\r\n      --version  shows software version\r\n");

   exit(1);
}
